#include "buffer.h"




/////////
/////////   Buffer Functions
/////////

Buffer::Buffer(int _width, int _height){
  width = _width;
  height = _height;

  itterations = 255;
  samples = 4;
  contrast = 2;
  fractal = 1;
  ratio = (float)width/(float)height;

  setCursor(0, 0, 1, 0);
  setStart(0, 0, 1, 0);
  setEnd(0, 0, 1, 0);

  buffer.resize(width, std::vector<int>(height));
}

void Buffer::clear(){
  maxValue = 0;
  for(int x=0; x<width; x++){
    for(int y=0; y<height; y++){
      buffer[x][y] = 0;
    }
  }
}

void Buffer::save(std::string _filename){
  FILE *f = fopen( _filename.c_str(), "w" );         // Write image to PPM file.
   fprintf(f, "P3\n%d %d\n%d\n", width, height, 255);
   for(int y=0; y<height; y++){
     for(int x=0; x<width; x++){

       float bright = map((float)buffer[x][y], 0, (float)maxValue, 0, 1.0);
       int value = (int)map(pow(bright, 1.0/contrast), 0, 1, 0, 255);

       fprintf(f,"%d %d %d ", value, value, value);
     }
   }
}




/////////
/////////   Buffer Settings
/////////

int Buffer::getMax(){
  return maxValue;
}

void Buffer::setItterations(int _itterations){
  itterations = _itterations;
}
void Buffer::setSamples(int _samples){
  samples = _samples;
}
void Buffer::setContrast(float _contrast){
  contrast = _contrast;
}

void Buffer::setFractal(unsigned char _fractal){
  fractal = _fractal;
}

std::string Buffer::stats(){
  std::ostringstream output;
  output << "Buffer: " << width << "x" << height;
  output << "    Samples: " << samples << "    Itterations: " << itterations;
  output << "    Contrast: " << contrast;
  return output.str();
}




/////////
/////////   Translate Coordinates
/////////

void Buffer::setCursor(float _x, float _y, float _s, float _r){
  cur.x = _x;
  cur.y = _y;
  cur.s = _s;
  cur.r = _r;
}
void Buffer::setStart(float _x, float _y, float _s, float _r){
  start.x = _x;
  start.y = _y;
  start.s = _s;
  start.r = _r;
}
void Buffer::setEnd(float _x, float _y, float _s, float _r){
  end.x = _x;
  end.y = _y;
  end.s = _s;
  end.r = _r;
}

void Buffer::cursorUpdate(float _f){
  cur.x = start.x + (end.x-start.x)*_f;
  cur.y = start.y + (end.y-start.y)*_f;
  cur.s = start.s + (end.s-start.s)*_f;
  cur.r = start.r + (end.r-start.r)*_f;
}

float Buffer::map(float _v, float _a1, float _a2, float _b1, float _b2){
  return _b1 + (_v-_a1)*(_b2-_b1)/(_a2-_a1);
}

Coords Buffer::translate(int _x, int _y){


  float angle = -cur.r * M_PI / 180.0;

  float rx = (float)rand()/(float)RAND_MAX;
  float ry = (float)rand()/(float)RAND_MAX;

  Coords org; // origin point
  org.x = ((((float)_x+rx) / (float)width * 2) - 1) * ratio; // X Origin
  org.y = (((float)_y+ry) / (float)height * 2) - 1;          // Y Origin

  float x = cos(angle)*org.x*cur.s - sin(angle)*org.y*cur.s + cur.x;  // New X
  float y = sin(angle)*org.x*cur.s + cos(angle)*org.y*cur.s + cur.y;  // New Y

  Coords tmpCords = {x, y};
  return tmpCords;
}

Pixel Buffer::translate(float _x, float _y){

  float angle = cur.r * M_PI / 180.0;

  Coords org; // origin point
  org.x = cos(angle)*(_x-cur.x)/cur.s - sin(angle)*(_y-cur.y)/cur.s;
  org.y = sin(angle)*(_x-cur.x)/cur.s + cos(angle)*(_y-cur.y)/cur.s;

  float x = (org.x/ratio + 1) / 2 * width;  // New X
  float y = (org.y + 1) / 2 * height;  // New Y

  Pixel tmpPix = {(int)x, (int)y};
  return tmpPix;
}




/////////
/////////   Render Fractal
/////////

void Buffer::render(int _x, int _y){
  Coords c = translate(_x, _y);

  float a = c.x;
  float b = c.y;

  int n = 0;
  float z = 0;
  while(n<=itterations){
    float aa = a*a - b*b;
    float bb = 2.0 * a * b;
    a = aa + c.x;
    b = bb + c.y;
    if(abs(a + b) > 4){ break; }

    if(fractal==2 || fractal==0){
      Pixel p = translate(a, b);
      if(p.x>=0 && p.x<width && p.y>=0 && p.y<height){
        buffer[p.x][p.y] += 1;
        if(buffer[p.x][p.y]>maxValue){ maxValue = buffer[p.x][p.y]; }
      }
    }

    n++;
  }
  if(fractal==1 || fractal==0){
    if(n >= itterations){ n=0; }
    buffer[_x][_y] += n;
    if(buffer[_x][_y] > maxValue){ maxValue = buffer[_x][_y]; }
  }
}
